package pps.core.common.utils;

import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUnit;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.lang.UUID;
import org.apache.commons.lang3.StringUtils;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.entity.BaseModel;
import pps.core.common.session.PpsUserSession;

import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * 封装工具类
 *
 * @author ZWT
 * @date 2023/08/25 16:08
 */
public class BaseUtils {

    /**
     * 按每500个一组分割
     */
    public static final Integer BATCH_SIZE = 500;

    /**
     * 计算批次数
     */
    public static Integer countStep(Integer totalSize) {
        return (totalSize + BATCH_SIZE - 1) / BATCH_SIZE;
    }

    /**
     * 分批处理
     *
     * @param list 列表
     * @return {@link List}<{@link List}<{@link E}>>
     */
    public static <E> List<List<E>> getSubList(List<E> list) {
        return Stream.iterate(0, n -> n + 1)
                .limit(countStep(list.size()))
                .parallel()
                .map(a -> list.stream().skip(a * BaseUtils.BATCH_SIZE)
                        .limit(BATCH_SIZE).parallel()
                        .collect(Collectors.toList()))
                .collect(Collectors.toList());
    }

    /**
     * 设置实体默认值
     *
     * @param model   实体
     * @param session 会话
     */
    public static void setBaseModelDefault(BaseModel model, PpsUserSession session) {
        String id = model.getId();
        Date now = new Date();
        String userId = session.getId();
        String userName = session.getUserName();
        if (StringUtils.isEmpty(id)) {
            model.setCreateById(userId);
            model.setCreateByName(userName);
            model.setCreateTime(now);
        }
        model.setModifyById(userId);
        model.setModifyByName(userName);
        model.setModifyTime(now);
    }

    /**
     * 设置实体默认值(定时任务)
     *
     * @param model 模型
     */
    public static void setBaseModelDefaultForJob(BaseModel model) {
        Date now = new Date();
        model.setId(BaseUtils.getUUID());
        model.setCreateById("Task");
        model.setCreateByName("Task");
        model.setCreateTime(now);
        model.setModifyById("Task");
        model.setModifyByName("Task");
        model.setModifyTime(now);
    }

    /**
     * 获取uuid
     *
     * @return {@link String}
     */
    public static String getUUID() {
        return StringUtils.replace(UUID.randomUUID().toString(), "-", "");
    }

    /**
     * 获取周执行周期
     *
     * @param date 日期
     * @return {@link String}
     */
    public static String getExecutionCycleForWeek(DateTime date) {
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append(date.year())
                .append("年第")
                .append(date.weekOfYear())
                .append('周');
        return stringBuilder.toString();
    }

    /**
     * 获取校准执行周期
     *
     * @param date   日期
     * @param offset 抵消
     * @return {@link String}
     */
    public static String getExecutionCycleForCalibration(Date date, int offset) {
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append(DateUtil.format(date, BusinessConstant.DATE_FORMAT_DAY_C))
                .append('-')
                .append(DateUtil.format(DateUtil.offsetDay(date, offset - 1), BusinessConstant.DATE_FORMAT_DAY_C));
        return stringBuilder.toString();
    }

    /**
     * 获取校准执行周期
     *
     * @param startDate 开始日期
     * @param endDate   结束日期
     * @return {@link String}
     */
    public static String getExecutionCycleForCalibration(Date startDate, Date endDate) {
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append(DateUtil.format(startDate, BusinessConstant.DATE_FORMAT_DAY_C))
                .append('-')
                .append(DateUtil.format(endDate, BusinessConstant.DATE_FORMAT_DAY_C));
        return stringBuilder.toString();
    }

    /**
     * 获取时差分钟
     *
     * @param optimizeTime 优化后时间
     * @param originalTime 原始时间
     * @return int
     */
    public static int getTimeDifferenceMinute(DateTime optimizeTime, DateTime originalTime) {
        int between = (int) optimizeTime.between(originalTime, DateUnit.MINUTE);
        if (optimizeTime.compareTo(originalTime) < 0) {
            between = ~between + 1;
        }
        return between;
    }

    /**
     * 获取结束时间字符串
     *
     * @param endTimeOptimize 结束时间优化
     * @return {@link String}
     */
    public static String getEndTimeString(DateTime endTimeOptimize) {
        String endTimeString;
        if (endTimeOptimize.compareTo(BusinessConstant.DATE_FLAG) > 0) {
            endTimeString = BusinessConstant.END_OF_DAY_TIME;
        } else {
            endTimeString = endTimeOptimize.toString(BusinessConstant.MINUTES_FORMAT);
        }
        return endTimeString;
    }

    /**
     * 获取月份字符串
     *
     * @param monthNum 月份
     * @return {@link String}
     */
    public static String getMonthString(int monthNum) {
        String month;
        if (monthNum < BusinessConstant.TEN) {
            month = "0" + monthNum;
        } else {
            month = String.valueOf(monthNum);
        }
        return month;
    }
}