package pps.core.base.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.IoUtil;
import cn.hutool.core.io.NioUtil;
import cn.hutool.core.io.file.FileReader;
import cn.hutool.core.text.CharSequenceUtil;
import cn.hutool.core.util.ArrayUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.sun.mail.imap.IMAPFolder;
import org.apache.commons.lang3.StringUtils;
import org.springframework.web.client.RestTemplate;
import pps.cloud.base.service.IBaseWeatherCloudService;
import pps.core.base.entity.BasePhotovoltaicPlantView;
import pps.core.base.entity.PlantPredictedPowerDataEnt;
import pps.core.base.enums.WindDirection;
import pps.core.base.mapper.BasePhotovoltaicPlantViewMapper;
import pps.core.base.mapper.PlantPredictedPowerDataMapper;
import pps.core.base.mapper.PlantPredictedPowerDataViewMapper;
import pps.core.base.service.data.DailyData;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.utils.BaseUtils;
import xstartup.annotation.XService;
import xstartup.base.XContext;
import xstartup.base.XStartup;
import xstartup.base.exception.XServiceException;
import xstartup.base.tool.XLoggerTool;
import xstartup.base.util.XCopyUtils;
import xstartup.core.base.helper.XThreadHelper;
import xstartup.data.XServiceResult;
import xstartup.helper.XTransactionHelper;

import javax.mail.*;
import javax.mail.internet.MimeUtility;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.time.LocalTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 天气数据接收Cloud模块
 *
 * @author ZWT
 * @date 2023/11/13 17:54
 */
@XService
public class BaseWeatherCloudServiceImpl implements IBaseWeatherCloudService {

    /**
     * 天气数据接收Cloud模块--天气数据处理
     *
     * @param context 上下文
     * @return {@link XServiceResult}
     */
    @Override
    public XServiceResult weatherDataProcess(XContext context) {
        XLoggerTool logger = context.getLogger();
        String host = XStartup.getCurrent().getProperty("weather.mail.host");
        String username = XStartup.getCurrent().getProperty("weather.mail.username");
        String password = XStartup.getCurrent().getProperty("weather.mail.password");
        String protocol = XStartup.getCurrent().getProperty("weather.mail.protocol");
        return XTransactionHelper.begin(context, () -> {
            //取未读邮件
            Store store = null;
            Folder folder = null;
            try {
                Properties props = System.getProperties();
                props.setProperty("mail.store.protocol", protocol);
                Session session = Session.getDefaultInstance(props, null);
                store = session.getStore(protocol);
                store.connect(host, username, password);
                folder = store.getFolder("INBOX");
                if (folder instanceof IMAPFolder) {
                    IMAPFolder imapFolder = (IMAPFolder) folder;
                    // javamail中使用id命令有校验checkOpened, 所以要去掉id方法中的checkOpened();
                    imapFolder.doCommand(p -> {
                        Map<String, String> gmap = new HashMap<>();
                        gmap.put("GUID", "FUTONG");
                        p.id(gmap);
                        return null;
                    });
                }
                if (Objects.nonNull(folder)) {
                    folder.open(Folder.READ_WRITE);
                }
                logger.info("------ weatherDataProcess connect mail ------");
                Message[] messages = folder.getMessages();
                if (ArrayUtil.isNotEmpty(messages)) {
                    List<Message> messageList = new ArrayList<>(16);
                    for (Message message : messages) {
                        boolean isNew = false;
                        for (Flags.Flag flag : message.getFlags().getSystemFlags()) {
                            if (flag == Flags.Flag.SEEN) {
                                isNew = true;
                            }
                        }
                        if (!isNew && CharSequenceUtil.endWith(message.getSubject(), "天气预测数据")) {
                            messageList.add(message);
                        }
                    }
                    if (CollUtil.isEmpty(messageList)) {
                        logger.info("------ weatherDataProcess messageList is empty ------");
                        return null;
                    }
                    //取电站
                    BasePhotovoltaicPlantViewMapper viewMapper = context.getBean(BasePhotovoltaicPlantViewMapper.class);
                    List<BasePhotovoltaicPlantView> plantList = viewMapper.selectPlantList();
                    if (CollUtil.isEmpty(plantList)) {
                        logger.info("------ weatherDataProcess plantList is empty ------");
                        return null;
                    }
                    //按照cityCode分组
                    Map<String, List<BasePhotovoltaicPlantView>> plantMap = plantList.stream()
                            .collect(Collectors.groupingBy(BasePhotovoltaicPlantView::getCityCode));
                    PlantPredictedPowerDataMapper mapper = context.getBean(PlantPredictedPowerDataMapper.class);
                    PlantPredictedPowerDataViewMapper dataViewMapper = context.getBean(PlantPredictedPowerDataViewMapper.class);
                    List<String> timeList = Arrays.asList("15:00", "30:00", "45:00");
                    //读取附件
                    for (Message message : messageList) {
                        String tempFilePath = this.saveAttachment(message);
                        logger.info("------ weatherDataProcess load file: {}", tempFilePath);
                        FileReader fileReader = new FileReader(tempFilePath);
                        String jsonString = fileReader.readString();
                        JSONObject jsonObject = JSON.parseObject(jsonString);
                        Map<String, List<DailyData>> cityDataMap = new HashMap<>(16);
                        for (Map.Entry<String, Object> entry : jsonObject.entrySet()) {
                            String key = entry.getKey();
                            Object value = entry.getValue();
                            if (value instanceof JSONArray) {
                                cityDataMap.put(key, ((JSONArray) value).toJavaList(DailyData.class));
                            }
                        }
                        for (String cityCode : cityDataMap.keySet()) {
                            if (!plantMap.containsKey(cityCode)) {
                                continue;
                            }
                            List<DailyData> dailyDataList = cityDataMap.get(cityCode);
                            List<BasePhotovoltaicPlantView> plantViewList = plantMap.get(cityCode);
                            //遍历预测数据列表
                            for (DailyData dailyData : dailyDataList) {
                                for (BasePhotovoltaicPlantView plant : plantViewList) {
                                    List<PlantPredictedPowerDataEnt> batchList = new ArrayList<>();
                                    List<PlantPredictedPowerDataEnt> allList = new ArrayList<>();
                                    List<PlantPredictedPowerDataEnt> objList = new ArrayList<>();
                                    DateTime dateTime = DateTime.of(dailyData.getDateTime());
                                    int hours = 0;
                                    String plantId = plant.getId();
                                    DateTime date;
                                    for (int i = 0; i < 8; i++) {
                                        //判断小时数与查到的小时数，如果查到的小时数小，判断为第二天
                                        int hour = LocalTime.parse(dailyData.getHourTimeArray().get(i)).getHour();
                                        if (hours >= hour) {
                                            dateTime = DateUtil.offsetDay(dateTime, 1);
                                        }
                                        hours = hour;
                                        PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                                this.int2Str(hours),
                                                new BigDecimal(CharSequenceUtil.replace(dailyData.getTemperatureArray().get(i), "℃", "")),
                                                new BigDecimal(CharSequenceUtil.replace(dailyData.getWindSpeedArray().get(i), "m/s", "")),
                                                new BigDecimal(Objects.requireNonNull(WindDirection.findByValue(dailyData.getWindDirectionArray().get(i))).getValue()),
                                                new BigDecimal(CharSequenceUtil.replace(dailyData.getPressureArray().get(i), "hPa", "")),
                                                new BigDecimal(CharSequenceUtil.replace(dailyData.getHumidityArray().get(i), "%", "")),
                                                String.valueOf(dateTime.year()),
                                                this.int2Str(dateTime.month() + 1),
                                                this.int2Str(dateTime.dayOfMonth())
                                        );
                                        // 5.执行SQL
                                        mapper.delete(new LambdaQueryWrapper<PlantPredictedPowerDataEnt>()
                                                .eq(PlantPredictedPowerDataEnt::getPlantId, plantId)
                                                .eq(PlantPredictedPowerDataEnt::getYearTime, dao.getYearTime())
                                                .eq(PlantPredictedPowerDataEnt::getMonthTime, dao.getMonthTime())
                                                .eq(PlantPredictedPowerDataEnt::getDayTime, dao.getDayTime())
                                                .eq(PlantPredictedPowerDataEnt::getHourTime, dao.getHourTime())
                                        );
                                        allList.add(dao);
                                        PlantPredictedPowerDataEnt isExit = objList.stream().filter(item ->
                                                        CharSequenceUtil.equals(item.getYearTime(), dao.getYearTime()) &&
                                                                CharSequenceUtil.equals(item.getMonthTime(), dao.getMonthTime()) &&
                                                                CharSequenceUtil.equals(item.getDayTime(), dao.getDayTime()) &&
                                                                CharSequenceUtil.equals(item.getHourTime(), dao.getHourTime()))
                                                .findFirst()
                                                .orElse(null);
                                        if (Objects.isNull(isExit)) {
                                            objList.add(dao);
                                        }
                                    }
                                    // 原站点时间结构为：08:00	11:00	14:00	17:00	20:00	23:00	02:00	05:00
                                    // 两条数据中间差了两个小时，使用上下两条数据推算出中间差掉的两个小时
                                    for (int i = 0; i < objList.size() - 1; i++) {
                                        for (int y = 1; y < 3; y++) {
                                            PlantPredictedPowerDataEnt dataEnt = objList.get(i);
                                            date = DateUtil.parseDate(dataEnt.getDataDate());
                                            String hourString;
                                            Integer hourTime = Integer.valueOf(dataEnt.getHourTime());
                                            if (hourTime.equals(23)) {
                                                date = DateUtil.offsetDay(date, 1);
                                                hourString = "0" + (y - 1);
                                            } else {
                                                if ((hourTime + y) < 10) {
                                                    hourString = "0" + (hourTime + y);
                                                } else {
                                                    hourString = String.valueOf(hourTime + y);
                                                }
                                            }
                                            PlantPredictedPowerDataEnt nextData = objList.get(i + 1);
                                            BigDecimal valueOf = BigDecimal.valueOf(0.3 * y);
                                            PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                                    hourString,
                                                    this.compute(dataEnt.getTemperature(), nextData.getTemperature(), valueOf),
                                                    this.compute(dataEnt.getWindSpeed(), nextData.getWindSpeed(), valueOf),
                                                    dataEnt.getWindDirection(),
                                                    this.compute(dataEnt.getPressure(), nextData.getPressure(), valueOf),
                                                    this.compute(dataEnt.getHumidity(), nextData.getHumidity(), valueOf),
                                                    String.valueOf(date.year()),
                                                    this.int2Str(date.month() + 1),
                                                    this.int2Str(date.dayOfMonth())
                                            );
                                            // 5.执行SQL
                                            mapper.delete(new LambdaQueryWrapper<PlantPredictedPowerDataEnt>()
                                                    .eq(PlantPredictedPowerDataEnt::getPlantId, plantId)
                                                    .eq(PlantPredictedPowerDataEnt::getYearTime, dao.getYearTime())
                                                    .eq(PlantPredictedPowerDataEnt::getMonthTime, dao.getMonthTime())
                                                    .eq(PlantPredictedPowerDataEnt::getDayTime, dao.getDayTime())
                                                    .eq(PlantPredictedPowerDataEnt::getHourTime, dao.getHourTime())
                                            );
                                            allList.add(dao);
                                        }
                                    }
                                    //入库
                                    List<PlantPredictedPowerDataEnt> sortList = allList.stream()
                                            .sorted(Comparator.comparing(PlantPredictedPowerDataEnt::getDataDate))
                                            .collect(Collectors.toList());
                                    for (int i = 0; i < sortList.size(); i++) {
                                        PlantPredictedPowerDataEnt item = sortList.get(i);
                                        batchList.add(item);
                                        if (i < sortList.size() - 1) {
                                            for (int y = 0; y < timeList.size(); y++) {
                                                PlantPredictedPowerDataEnt dao = XCopyUtils.copyNewObject(item);
                                                dao.setMinTime(timeList.get(y));
                                                dao.setDataDate(dao.getYearTime() + '-' + dao.getMonthTime() + '-' + dao.getDayTime() + ' ' + dao.getHourTime() + ':' + dao.getMinTime());
                                                PlantPredictedPowerDataEnt nextData = sortList.get(i + 1);
                                                BigDecimal valueOf = BigDecimal.valueOf(0.3 * (y + 1));
                                                dao.setTemperature(this.compute(Objects.isNull(item.getTemperature()) ? BigDecimal.ZERO : item.getTemperature(),
                                                        Objects.isNull(nextData.getTemperature()) ? BigDecimal.ZERO : nextData.getTemperature(),
                                                        valueOf));
                                                dao.setHumidity(this.compute(Objects.isNull(item.getHumidity()) ? BigDecimal.ZERO : item.getHumidity(),
                                                        Objects.isNull(nextData.getHumidity()) ? BigDecimal.ZERO : nextData.getHumidity(),
                                                        valueOf));
                                                dao.setWindSpeed(this.compute(Objects.isNull(item.getWindSpeed()) ? BigDecimal.ZERO : item.getWindSpeed(),
                                                        Objects.isNull(nextData.getWindSpeed()) ? BigDecimal.ZERO : nextData.getWindSpeed(),
                                                        valueOf));
                                                dao.setPressure(this.compute(Objects.isNull(item.getPressure()) ? BigDecimal.ZERO : item.getPressure(),
                                                        Objects.isNull(nextData.getPressure()) ? BigDecimal.ZERO : nextData.getPressure(),
                                                        valueOf));
                                                dao.setPlaneIrradiance(this.compute(Objects.isNull(item.getPlaneIrradiance()) ? BigDecimal.ZERO : item.getPlaneIrradiance(),
                                                        Objects.isNull(nextData.getPlaneIrradiance()) ? BigDecimal.ZERO : nextData.getPlaneIrradiance(),
                                                        valueOf));
                                                dao.setHorizontalIrradiance(this.compute(Objects.isNull(item.getHorizontalIrradiance()) ? BigDecimal.ZERO : item.getHorizontalIrradiance(),
                                                        Objects.isNull(nextData.getHorizontalIrradiance()) ? BigDecimal.ZERO : nextData.getHorizontalIrradiance(),
                                                        valueOf));
                                                dao.setPower(this.compute(Objects.isNull(item.getPower()) ? BigDecimal.ZERO : item.getPower(),
                                                        Objects.isNull(nextData.getPower()) ? BigDecimal.ZERO : nextData.getPower(),
                                                        valueOf));
                                                batchList.add(dao);
                                            }
                                        }
                                    }
                                    if (CollUtil.isNotEmpty(batchList)) {
                                        int size = batchList.size();
                                        logger.info("------ weatherDataProcess batchList insert plantId: {} size: {}", plantId, size);
                                        if (size > BaseUtils.BATCH_SIZE) {
                                            List<List<PlantPredictedPowerDataEnt>> subList = BaseUtils.getSubList(batchList);
                                            subList.forEach(dataViewMapper::insertBatch);
                                        } else {
                                            dataViewMapper.insertBatch(batchList);
                                        }
                                    }
                                    //训练接口http://127.0.0.1:10098/aiprediction/xgbtrain?plantId=018a64635ac47cf58895147b0e1bf7e3
                                    //自动调用预测接口
                                    XThreadHelper.async(() -> {
                                        try {
                                            Thread.sleep(3000);
                                            logger.info("------ weatherDataProcess http prediction plantId: {} ", plantId);
                                            RestTemplate restTemplate = context.getBean(RestTemplate.class);
                                            String forObject = restTemplate.getForObject("http://127.0.0.1:10098/aiprediction/xgbreason?plantId=" + plantId, String.class);
                                            logger.info("------ weatherDataProcess http prediction res: {} ", forObject);
                                        } catch (Exception e) {
                                            logger.error("------ weatherDataProcess Exception", e);
                                            throw new XServiceException(e);
                                        }
                                    });
                                }
                            }
                        }
                        logger.info("------ weatherDataProcess del temp file path: {}", tempFilePath);
                        FileUtil.del(tempFilePath);
                    }
                }
            } catch (MessagingException e) {
                logger.error("------ weatherDataProcess MessagingException", e);
                throw new XServiceException(e);
            } catch (IOException e) {
                logger.error("------ weatherDataProcess IOException", e);
                throw new XServiceException(e);
            } finally {
                NioUtil.close(folder);
                NioUtil.close(store);
            }
            return XServiceResult.OK;
        });
    }

    /**
     * 获取电厂预测功率ent
     *
     * @param plantId       电站ID
     * @param hourTime      小时
     * @param temperature   温度
     * @param windSpeed     风速
     * @param windDirection 风向
     * @param pressure      压力
     * @param humidity      湿度
     * @param yearTime      年时间
     * @param monthTime     月份时间
     * @param dayTime       白天
     * @return {@link PlantPredictedPowerDataEnt}
     */
    private PlantPredictedPowerDataEnt getPlantPredictedPowerEnt(String plantId, String hourTime, BigDecimal temperature, BigDecimal windSpeed, BigDecimal windDirection,
                                                                 BigDecimal pressure, BigDecimal humidity, String yearTime, String monthTime, String dayTime) {
        PlantPredictedPowerDataEnt dao = new PlantPredictedPowerDataEnt();
        dao.setCreateTime(new Date());
        dao.setPlantId(plantId);
        dao.setHourTime(hourTime);
        dao.setTemperature(temperature);
        dao.setWindSpeed(windSpeed);
        dao.setWindDirection(windDirection);
        dao.setPressure(pressure);
        dao.setHumidity(humidity);
        dao.setYearTime(yearTime);
        dao.setMonthTime(monthTime);
        dao.setDayTime(dayTime);
        dao.setMinTime(BusinessConstant.START_OF_DAY_TIME);
        dao.setDataDate(dao.getYearTime() + '-' + dao.getMonthTime() + '-' + dao.getDayTime() + ' ' + dao.getHourTime() + ':' + dao.getMinTime());
        return dao;
    }


    /**
     * 计算
     *
     * @param a 一
     * @param b b
     * @param c c
     * @return {@link BigDecimal}
     */
    private BigDecimal compute(BigDecimal a, BigDecimal b, BigDecimal c) {
        return a.add(b.subtract(a).multiply(c));
    }

    /**
     * int转String
     *
     * @param num 数字
     * @return {@link String}
     */
    private String int2Str(int num) {
        return num < 10 ? "0" + num : String.valueOf(num);
    }


    /**
     * 保存附件
     *
     * @param part 部分
     * @return {@link String}
     * @throws MessagingException 消息传递异常
     * @throws IOException        IOException
     */
    private String saveAttachment(Part part) throws MessagingException, IOException {
        String fileName;
        String filePath = null;
        if (part.isMimeType("multipart/*")) {
            Multipart mp = (Multipart) part.getContent();
            BodyPart mPart;
            for (int i = 0; i < mp.getCount(); i++) {
                mPart = mp.getBodyPart(i);
                String disposition = mPart.getDisposition();
                if (Part.ATTACHMENT.equals(disposition) || Part.INLINE.equals(disposition)) {
                    fileName = mPart.getFileName();
                    if (CharSequenceUtil.isNotEmpty(fileName) && StringUtils.lowerCase(fileName).contains("GB2312")) {
                        fileName = MimeUtility.decodeText(fileName);
                    }
                    fileName = CharSequenceUtil.replace(fileName, "\\?|=", "");
                    filePath = this.saveTempFile(fileName, mPart.getInputStream());
                } else if (mPart.isMimeType("multipart/*")) {
                    this.saveAttachment(mPart);
                } else {
                    fileName = mPart.getFileName();
                    if (CharSequenceUtil.isNotEmpty(fileName) && StringUtils.lowerCase(fileName).contains("GB2312")) {
                        fileName = MimeUtility.decodeText(fileName);
                        filePath = this.saveTempFile(fileName, mPart.getInputStream());
                    }
                }
            }
        } else if (part.isMimeType("message/rfc822")) {
            this.saveAttachment((Part) part.getContent());
        }
        return filePath;
    }

    /**
     * 保存临时文件
     *
     * @param fileName 文件名
     * @param in       在里面
     * @return {@link String}
     */
    private String saveTempFile(String fileName, InputStream in) {
        String storeDir = XStartup.getCurrent().getProperty("weather.file.temp.path");
        StringBuilder stringBuilder = new StringBuilder(storeDir)
                .append(File.separator)
                .append(DateTime.now().toString("yyyyMMdd"));
        File savePath = new File(stringBuilder.toString());
        if (!savePath.exists()) {
            savePath.mkdirs();
        }
        String filePath = stringBuilder.append(File.separator).append(fileName).toString();
        FileUtil.writeFromStream(in, filePath);
        IoUtil.close(in);
        return filePath;
    }
}
