package pps.core.space.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import pps.cloud.prediction.service.data.plant_predicted_power_data.DynamicQueryPlantPredictedPowerInput;
import pps.cloud.prediction.service.data.storage_predicted_power_data.DynamicQueryStoragePredictedPowerInput;
import pps.cloud.space.service.ISpaceOptimizeMidCloudService;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.utils.BaseUtils;
import pps.core.space.entity.SpaceInstitutionDetailEnt;
import pps.core.space.entity.SpaceOptimizeDurationDTO;
import pps.core.space.entity.SpaceOptimizePeriodDTO;
import pps.core.space.entity.SpaceOptimizeWellheadDTO;
import pps.core.space.mapper.SpaceOptimizeViewMapper;
import pps.core.space.service.data.SpaceOptimizeLineRelation;
import xstartup.annotation.XService;
import xstartup.base.XContext;
import xstartup.data.XServiceResult;
import xstartup.helper.XTransactionHelper;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * 中短期间开优化Cloud模块
 *
 * @author ZWT
 * @date 2023/09/18 14:14
 */
@XService
public class SpaceOptimizeMidCloudServiceImpl extends SpaceOptimizeBaseService implements ISpaceOptimizeMidCloudService {

    /**
     * 中短期间开优化Cloud模块--定时任务
     *
     * @param context 上下文
     * @return {@link XServiceResult}
     */
    @Override
    public XServiceResult optimizeMidJob(XContext context) {
        //取当前时间
        DateTime date = DateUtil.date();
        String startWeek = date.toString(BusinessConstant.DATE_FORMAT_DAY);
        //取生效中的基础间开
        List<SpaceInstitutionDetailEnt> detailEntList = super.getEffectiveSpaceInstitution(context, startWeek);
        SpaceOptimizeLineRelation relation = super.getOptimizeLineRelation(context, detailEntList);
        //月份要加一
        int monthNum = date.month() + 1;
        DateTime lastWeek = DateUtil.lastWeek();
        List<SpaceOptimizePeriodDTO> periodDTOList = new ArrayList<>(32);
        List<SpaceOptimizeWellheadDTO> wellheadDTOList = new ArrayList<>(64);
        List<SpaceOptimizeDurationDTO> durationDTOList = new ArrayList<>(128);
        List<SpaceOptimizeDurationDTO> unOptimizeDurationList = new ArrayList<>(128);
        String executionCycleForWeek = BaseUtils.getExecutionCycleForWeek(date);
        //日期查询范围结束条件,下周一
        String nextWeek = DateUtil.beginOfWeek(DateUtil.nextWeek()).toString(BusinessConstant.DATE_FORMAT_DAY);
        //优化
        for (SpaceInstitutionDetailEnt detail : detailEntList) {
            String detailId = detail.getId();
            String lineId = detail.getLineId();
            Date institutionEndDate = detail.getInstitutionEndDate();
            //优化截止日期
            DateTime optimizeDeadline = DateUtil.endOfWeek(date);
            if (optimizeDeadline.compareTo(institutionEndDate) > 0) {
                optimizeDeadline = DateUtil.endOfDay(institutionEndDate);
            }
            //创建记录
            String periodId = super.createOptimizePeriod(periodDTOList, detailId, lineId, executionCycleForWeek, optimizeDeadline);
            switch (detail.getGridTypeKey()) {
                //并网型优化
                case "1":
                    super.gridConnectedOptimization(context, wellheadDTOList, durationDTOList, unOptimizeDurationList,
                            relation.getWellheadsByInstitutionIdIdMap().get(detail.getId()), detail, monthNum, periodId,
                            DynamicQueryPlantPredictedPowerInput.builder()
                                    .plantIds(relation.getPlantIdsByLineIdMap().get(lineId))
                                    .startTime(startWeek)
                                    .endTime(nextWeek)
                                    .dateType(BusinessConstant.ONE)
                                    .build()
                    );
                    break;
                //离网型优化
                case "0":
                    super.offGridOptimization(context, wellheadDTOList, durationDTOList, unOptimizeDurationList,
                            relation.getWellheadsByInstitutionIdIdMap().get(detail.getId()), relation.getStorageAvgMap(), detail, periodId,
                            DynamicQueryStoragePredictedPowerInput.builder()
                                    .storageIds(relation.getStorageIdsByLineIdMap().get(detail.getLineId()))
                                    .startTime(DateUtil.beginOfWeek(lastWeek).toString(BusinessConstant.DATE_FORMAT_DAY))
                                    .endTime(DateUtil.endOfWeek(lastWeek).toString(BusinessConstant.DATE_FORMAT_DAY))
                                    .build(),
                            DynamicQueryPlantPredictedPowerInput.builder()
                                    .plantIds(relation.getPlantIdsByLineIdMap().get(lineId))
                                    .startTime(startWeek)
                                    .endTime(nextWeek)
                                    .dateType(BusinessConstant.ONE)
                                    .build());
                    break;
                default:
                    //电网类型不存在
            }
        }
        //开启事务
        return XTransactionHelper.begin(context, () -> {
            int size;
            SpaceOptimizeViewMapper optimizeViewMapper = context.getBean(SpaceOptimizeViewMapper.class);
            if (CollUtil.isNotEmpty(periodDTOList)) {
                size = periodDTOList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<SpaceOptimizePeriodDTO>> subList = BaseUtils.getSubList(periodDTOList);
                    subList.forEach(optimizeViewMapper::midPeriodBatchInsertList);
                } else {
                    optimizeViewMapper.midPeriodBatchInsertList(periodDTOList);
                }
            }
            if (CollUtil.isNotEmpty(wellheadDTOList)) {
                size = wellheadDTOList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<SpaceOptimizeWellheadDTO>> subList = BaseUtils.getSubList(wellheadDTOList);
                    subList.forEach(optimizeViewMapper::midWellheadBatchInsertList);
                } else {
                    optimizeViewMapper.midWellheadBatchInsertList(wellheadDTOList);
                }
            }
            if (CollUtil.isNotEmpty(unOptimizeDurationList)) {
                size = unOptimizeDurationList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<SpaceOptimizeDurationDTO>> subList = BaseUtils.getSubList(unOptimizeDurationList);
                    subList.forEach(optimizeViewMapper::midDurationBatchInsertList);
                } else {
                    optimizeViewMapper.midDurationBatchInsertList(unOptimizeDurationList);
                }
            }
            if (CollUtil.isNotEmpty(durationDTOList)) {
                size = durationDTOList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<SpaceOptimizeDurationDTO>> subList = BaseUtils.getSubList(durationDTOList);
                    subList.forEach(optimizeViewMapper::midDurationBatchInsertList);
                } else {
                    optimizeViewMapper.midDurationBatchInsertList(durationDTOList);
                }
            }
            return XServiceResult.OK;
        });
    }

    /*-----------------------------------private-----------------------------------*/
}