package pps.core.prediction.service;

import cn.hutool.core.collection.CollUtil;
import pps.cloud.base.service.IBasePowerLineCloudService;
import pps.cloud.base.service.data.base_photovoltaic_plant.GetBasePhotovoltaicPlantCloudInput;
import pps.cloud.base.service.data.base_photovoltaic_plant.GetBasePhotovoltaicPlantCloudOutput;
import pps.cloud.base.service.data.base_power_line_plant.DynamicQueryBasePowerLinePlantInput;
import pps.cloud.base.service.data.base_power_line_plant.DynamicQueryBasePowerLinePlantOutput;
import pps.core.common.constant.BusinessConstant;
import pps.core.prediction.service.data.energy_consumption_curve.GetEnergyConsumptionCurveInput;
import pps.core.prediction.service.data.energy_consumption_curve.GetEnergyConsumptionCurveOutput;
import xstartup.annotation.XService;
import xstartup.annotation.XText;
import xstartup.base.XContext;
import xstartup.base.util.XCopyUtils;
import xstartup.data.XListResult;
import xstartup.feature.api.annotation.XApiAnonymous;
import xstartup.feature.api.annotation.XApiGet;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * 用能曲线模块(测试用)
 * todo: 模拟测试用,后续替换
 *
 * @author ZWT
 * @date 2023/09/27 15:57
 */
@XText("用能曲线模块")
@XService
public class EnergyConsumptionCurveService extends StatisticAnalysisBaseService {

    /**
     * 用能曲线--发电列表
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XListResult}<{@link GetEnergyConsumptionCurveOutput}>
     */
    @XText("用能曲线--发电列表")
    @XApiAnonymous
    @XApiGet
    public XListResult<GetEnergyConsumptionCurveOutput> queryEnergyConsumptionCurveList(XContext context, GetEnergyConsumptionCurveInput input) {
        //通过线路ID查光伏电站
        List<DynamicQueryBasePowerLinePlantOutput> plantListByLineIds = this.getPowerLinePlantListByLineIds(context, CollUtil.newArrayList(input.getLineId()));
        List powerGenerationList = null;
        List<String> plantIdList = null;
        if (CollUtil.isNotEmpty(plantListByLineIds)) {
            plantIdList = plantListByLineIds.stream()
                    .map(DynamicQueryBasePowerLinePlantOutput::getPlantId)
                    .collect(Collectors.toList());
            powerGenerationList = super.getPhotovoltaicPowerGeneration(context, input.getDateType(), plantIdList, input.getStartTime(), input.getEndTime());
        }
        List<GetEnergyConsumptionCurveOutput> outputList = new ArrayList<>(0);
        if (CollUtil.isNotEmpty(powerGenerationList)) {
            XCopyUtils.copyList(powerGenerationList, outputList, GetEnergyConsumptionCurveOutput.class);
            //有发电量说明有电站ID,查电站详情
            List<GetBasePhotovoltaicPlantCloudOutput> plantList = super.getPhotovoltaicPlantByParam(context,
                    GetBasePhotovoltaicPlantCloudInput.builder()
                            .plantIds(plantIdList)
                            .build()
            );
            Optional<BigDecimal> reduce = plantList.stream()
                    .map(GetBasePhotovoltaicPlantCloudOutput::getTotalPower)
                    .reduce(BigDecimal::add);
            if (reduce.isPresent()) {
                //负荷率=实际发电量/理论最大发电量x100%
                BigDecimal maximumPowerGeneration = reduce.get();
                BigDecimal power;
                BigDecimal oneHundred = new BigDecimal(100);
                for (GetEnergyConsumptionCurveOutput output : outputList) {
                    power = output.getPower();
                    if (Objects.isNull(power) ||
                            power.compareTo(BigDecimal.ZERO) <= BusinessConstant.ZERO ||
                            maximumPowerGeneration.compareTo(BigDecimal.ZERO) == BusinessConstant.ZERO
                    ) {
                        output.setProductionLoad(BigDecimal.ZERO);
                        continue;
                    }
                    //计算负荷
                    output.setProductionLoad(
                            power.divide(maximumPowerGeneration, 4, RoundingMode.HALF_UP)
                                    .multiply(oneHundred)
                    );
                }
            }
        }
        return XListResult.success(outputList);
    }

    /*-----------------------------------private-----------------------------------*/

    /**
     * 按线路ID集合获取电站列表
     *
     * @param context 上下文
     * @param lineIds 线路ID
     * @return {@link List}<{@link DynamicQueryBasePowerLinePlantOutput}>
     */
    private List<DynamicQueryBasePowerLinePlantOutput> getPowerLinePlantListByLineIds(XContext context, List<String> lineIds) {
        IBasePowerLineCloudService cloudService = context.getBean(IBasePowerLineCloudService.class);
        XListResult<DynamicQueryBasePowerLinePlantOutput> result = cloudService.queryPowerLinePlantListByParam(context,
                DynamicQueryBasePowerLinePlantInput.builder()
                        .lineIds(lineIds)
                        .build()
        );
        result.throwIfFail();
        return result.getResult();
    }
}
